#!/bin/csh
#
# Copyright (C) 1997, Board of Trustees of the University of Illinois.
#
# Permission is granted to copy and distribute source with out fee.
# Commercialization of this product requires prior licensing from the
# University of Illinois.  Commercialization includes the integration of this 
# code in part or whole into a product for resale.  Free distribution of 
# unmodified source and use of this software is not considered 
# commercialization.

# user defined paratmers to control DLM Version 3
umask 022
set w_obj=1
set w_constr=1
set max_iter_default=1000000
set max_iter_g=50000000
set max_iter_ai=5000000
set max_iter_par=200000000
set max_iter_hanoi=50000000
set runs=10
set tabu_length=50
set flat_limit=50
set reset_c=10000
set reset_r=1.5

# nothing should be changed below this point
set res_dir=Results/v3
set res_sum=Results/result_sum_v3
if (! -d Results) mkdir Results
if (! -d $res_dir) mkdir $res_dir

switch ($1)
case "clean":
	rm -f $res_dir/*
	exit 0
case "ai":
case "as":
case "f":
case "g":
case "han":
case "ii":
case "jnh":
case "par":
case "ss":
case "tm":
	set list=$1
	breaksw
default:
	set list="ai as f g han ii jnh par ss tm"
	breaksw
endsw

foreach i (`echo $list`) 		# execute all benchmark problems
	switch ($i)
	case "g":
		set max_iter=$max_iter_g
		breaksw
	case "ai":
		set max_iter=$max_iter_ai
		breaksw
	case "par":
		set max_iter=$max_iter_par
		breaksw
	case "hanoi":
		set max_iter=$max_iter_hanoi
		breaksw
	default:
		set max_iter=$max_iter_default
		breaksw
	endsw

	set dir_list=`ls -1 DIMACS/$i*`
	foreach j ($dir_list)
		set k=`echo $j | sed -e "/cnf.gz/ s//r10/" -e "/DIMACS\// s///"`
		echo "Solving $k ..."
		zcat < $j | Bin/lgrg_solve_v3 $w_obj $w_constr $max_iter $runs \
		$tabu_length $flat_limit $reset_c $reset_r >! $res_dir/$k
	end
end

# print summary results in result_sum_v3 file

echo "" >! $res_sum
echo "Run of DLM Version 3 using the following parameters:" >> $res_sum
echo "" >> $res_sum
echo "    Weight of objective                    = $w_obj" >> $res_sum
echo "    Weight of constraints                  = $w_constr" >> $res_sum
echo "    max. iterations                        = $max_iter ($max_iter_ai for aim problems)" >> $res_sum
echo "    number of runs                         = $runs" >> $res_sum
echo "    length of tabu list                    = $tabu_length" >> $res_sum
echo "    flat move limit                        = $flat_limit" >> $res_sum
echo "    reset interval of Lagrange multipliers = $reset_c" >> $res_sum
echo "    reset rate of Lagrange multipliers     = $reset_r" >> $res_sum
echo "" >> $res_sum
echo ""|gawk '{printf("%20s\tsuccess\t\t\t  time/success (sec)\t   Iterations\n", "problem id")}'  >> $res_sum
echo ""|gawk '{printf("%20s\t\t\t      ave        min        max          ave\n", " ")}' >> $res_sum
echo "" >> $res_sum
set all_list=`ls -1 $res_dir/*`
foreach ii ($all_list)
	echo $ii | gawk '{kk=$ii; sub(".*/", "", kk); printf("%20s\t",kk)}' >> $res_sum
	grep Best $ii |gawk '{printf("(%s)\t\t", $4)}' >> $res_sum
	tail -4 $ii |grep 'Time/suc'| sed 's/,/ /g'| \
		gawk '{printf ("%10.3f %10.3f %10.3f",$5,$10,$12)}' >> $res_sum
	tail -4 $ii |grep 'Iter'| sed 's/,/ /g'| gawk '{printf ("%12.0f",$4)}' >> $res_sum
	echo "" >> $res_sum
end
