#!/bin/csh
#
# Copyright (C) 1997, Board of Trustees of the University of Illinois.
#
# Permission is granted to copy and distribute source with out fee.
# Commercialization of this product requires prior licensing from the
# University of Illinois.  Commercialization includes the integration of this 
# code in part or whole into a product for resale.  Free distribution of 
# unmodified source and use of this software is not considered 
# commercialization.
#
# DLM Version 3 with parameters set to regenerate the results in paper
# J60 located at http://manip.crhc.uiuc.edu/paper.html#NONLINEARDIS

# user defined paratmers to control DLM Version 3
umask 022
set w_obj_default=1
set w_constr_default=1
set max_iter_default=10000000
set runs=10
set tabu_length_default=50
set flat_limit=50
set reset_c=10000
set reset_r=1.5

# nothing should be changed below this point
set res_dir=Results/v3_paper
set res_sum=Results/result_sum_v3_paper
if (! -d Results) mkdir Results
if (! -d $res_dir) mkdir $res_dir

if ($1 == "clean") then
	rm -f $res_dir/*
	exit 0
else
	set list="f g125-17 g125-18 g250-15 g250-29 hanoi4 par8 par16-1.c par16-2.c par16-3.c par16-4.c par16-5.c par16-1-c par16-2-c par16-3-c par16-4-c par16-5-c"
endif

echo "" >! $res_sum
echo "Run of DLM Version 3 using the following parameters:" >> $res_sum
echo "" >> $res_sum

echo "    number of runs                         = $runs" >> $res_sum
echo "    flat move limit                        = $flat_limit" >> $res_sum
echo "    reset interval of Lagrange multipliers = $reset_c" >> $res_sum
echo "    reset rate of Lagrange multipliers     = $reset_r" >> $res_sum
echo "" >> $res_sum
echo "    Problem      w_obj      w_constr     max_iter     tabu_length" >> $res_sum

foreach i (`echo $list`) 		# execute all benchmark problems
	switch ($i)
	case "g125-17":
	case "g250-29":
		set w_obj=16
		set w_constr=$w_constr_default
		set max_iter=50000000
		set tabu_length=$tabu_length_default
		breaksw
	case "g125-18":
	case "g250-15":
		set w_obj=$w_obj_default
		set w_constr=2
		set max_iter=$max_iter_default
		set tabu_length=$tabu_length_default
		breaksw
	case "f":
		set w_obj=16
		set w_constr=$w_constr_default
		set max_iter=100000000
		set tabu_length=$tabu_length_default
		breaksw
	case "par8":
		set w_obj=$w_obj_default
		set w_constr=2
		set max_iter=10000000
		set tabu_length=$tabu_length_default
		breaksw
	case "par16-1.c":
	case "par16-2.c":
	case "par16-3.c":
	case "par16-4.c":
	case "par16-5.c":
		set w_obj=$w_obj_default
		set w_constr=$w_constr_default
		set max_iter=200000000
		set tabu_length=100
		breaksw
	case "par16-1-c":
	case "par16-2-c":
	case "par16-3-c":
	case "par16-4-c":
	case "par16-5-c":
		set w_obj=2
		set w_constr=$w_constr_default
		set max_iter=200000000
		set tabu_length=$tabu_length_default
		breaksw
	case "hanoi4":
		set w_obj=2
		set w_constr=$w_constr_default
		set max_iter=50000000
		set tabu_length=50
		breaksw
	endsw
	echo $i $w_obj $w_constr $max_iter $tabu_length | gawk '{printf("    %-10s %6d %12d %14d %13d\n", $1, $2, $3, $4, $5)}' >> $res_sum

	set dir_list=`ls -1 DIMACS/$i*`
	foreach j ($dir_list)
		set k=`echo $j | sed -e "/cnf.gz/ s//r10/" -e "/DIMACS\// s///"`
		echo "Solving $k ..."
		zcat < $j | Bin/lgrg_solve_v3 $w_obj $w_constr $max_iter $runs \
		$tabu_length $flat_limit $reset_c $reset_r >! $res_dir/$k
	end
end

# print summary results in result_sum_v3 file

echo "" >> $res_sum
echo ""|gawk '{printf("%20s\tsuccess\t\t\t  time/success (sec)\t   Iterations\n", "problem id")}'  >> $res_sum
echo ""|gawk '{printf("%20s\t\t\t      ave        min        max          ave\n", " ")}' >> $res_sum
echo "" >> $res_sum
set all_list=`ls -1 $res_dir/*`
foreach ii ($all_list)
	echo $ii | gawk '{kk=$ii; sub(".*/", "", kk); printf("%20s\t",kk)}' >> $res_sum
	grep Best $ii |gawk '{printf("(%s)\t\t", $4)}' >> $res_sum
	tail -4 $ii |grep 'Time/suc'| sed 's/,/ /g'| \
		gawk '{printf ("%10.3f %10.3f %10.3f",$5,$10,$12)}' >> $res_sum
	tail -4 $ii |grep 'Iter'| sed 's/,/ /g'| gawk '{printf ("%12.0f",$4)}' >> $res_sum
	echo "" >> $res_sum
end
