/*
 * Copyright (C) 1997, Board of Trustees of the University of Illinois.
 *
 * Permission is granted to copy and distribute source with out fee.
 * Commercialization of this product requires prior licensing from the
 * University of Illinois.  Commercialization includes the integration of this 
 * code in part or whole into a product for resale.  Free distribution of 
 * unmodified source and use of this software is not considered 
 * commercialization.
 *
 */

#include "lgrg_v2.h"

int main(int argc, char *argv[])
{
  lgrgrec *lgrg;
  FILE *fp; 

  printf("%s\n",copyright);

  /*
    if (argc != 2 && argc != 1) {
    printf("Usage: lgrg  inputfile.cnf or lgrg < inputfile.cnf\n");
    exit(-1);
    }

    if (argc == 2 ) {
    fprintf(stderr,"reading cnf file %s\n", argv[1]);
    fp = fopen(argv[1],"r");
    lgrg = lgrg_init(fp, TRUE);
    } else {
    fprintf(stderr,"reading cnf file from standard input\n");
    lgrg = lgrg_init(stdin, TRUE);
  }

  if (lgrg == NULL) {
    fprintf(stderr,"...reading cnf file failed\n");
    exit(-1);
  }  
  fclose(fp);
  */

  if (argc != 5) {
    printf("Usage: lgrg w_obj w_constr max_iter #runs < inputfile.cnf

  Input arguments:
  w_obj    -- integer (>= 0, e.g. 1), weight of objective
  w_constr -- integer(>= 1, e.g. 1),  weight of constraints
  max_iter -- integer (>=1, e.g. 500000), maximum iterations to run
  #runs    -- integer (>=1, e.g. 1, or 10), the number of runs
           if #runs=1, the initial point is origin, 0;
           if #runs>1, initial points are randomly generated\n");
    exit(-1);
  }

  INIT_U = atoi(argv[1]);
  INC_U  = atoi(argv[2]);
  NITER  = atoi(argv[3]);
  NROUND = atoi(argv[4]);

  printf("Weight of objective                     : %d
Weight of constraints                   : %d
Maximum number of iterations            : %d
Number of runs                          : %d\n",
	 INIT_U, INC_U, NITER, NROUND);
  if (NROUND==1) 
      printf("   (Starting from origin, 0)\n");
  else
      printf("   (Starting from randomly generated initial points)\n");

  lgrg = lgrg_init(stdin, FALSE);

  if (lgrg == NULL) {
    fprintf(stderr,"...reading cnf file failed\n");
    exit(-1);
  }  

  srand48(SEED);

  multi_start_rand(lgrg);
  
}

/* multi-start from randomly selected initial points */
void multi_start_rand(lgrgrec *lgrg)
{ 
  int iter, iterations, i, udx, udu, round, count;
  int encumbent, enci;
  REAL cur_obj, best_obj;
  struct tms t_buffer, old_t_buffer;
  int suc_time_s, suc_time_u, min_time_u, max_time_u, time_u, time_s;
  int aveevals = 0;
  int aveiter, miniter, maxiter, avefps;
  int avevalue = 0;

  miniter = min_time_u = lgrg->infinite;
  aveiter = maxiter = max_time_u = 0;

  times(&old_t_buffer);
  
  best_obj = lgrg->infinite;

  /* main body loop */
  avefps = 0;
  for (count = suc_time_s = suc_time_u =  round=0; round < NROUND; round++) {

    NEVALS = 0; 
    encumbent = 100000;
    lgrg->enable_wrong_cl_list=0;
    lgrg->diff_flag=0;
    diff_wc_flag = 0;

    fprintf(stdout, "\nRound %d\n", round);

    iterations = NITER;
    for (iter=0; iter<iterations; iter++) {
/*
      printf("iter %d, wrong cl %d, cur_lgrg %d, enc %d\n", iter, lgrg->wrong_clauses,lgrg->cur_lgrg, encumbent); 
*/

#ifdef DEBUG
/*      printf("iter %d, wrong cl %d, ture wrong cl %d, cur_lgrg %d, ture lgrg %d\n", iter, lgrg->wrong_clauses, eval_wrong_clauses(lgrg), lgrg->cur_lgrg, eval_lagrangian(lgrg)); */
      printf("iter %d, wrong cl %d, cur_lgrg %d\n", iter, lgrg->wrong_clauses,lgrg->cur_lgrg); 
#endif
      udx = update_x(lgrg);

      if(!(lgrg->wrong_clauses)){
#ifdef DEBUG
	check_lgrg(lgrg); 
#endif
	fprintf(stdout, "Lagrangian method finds feasible solution in %d iterations.\n", iter);
	count ++;
	best_obj = 0;

	times(&t_buffer);
	time_u = t_buffer.tms_utime - old_t_buffer.tms_utime;
	time_s = t_buffer.tms_stime - old_t_buffer.tms_stime;
	suc_time_s += time_s;
	suc_time_u += time_u;
	aveevals += NEVALS;
	aveiter += iter;
	avefps += iter*60*1.0/time_u;

	if (iter < miniter)
	  miniter = iter;
	if (iter > maxiter)
	  maxiter = iter;
	if (min_time_u > time_u)
	  min_time_u = time_u;
	if (max_time_u < time_u)
	  max_time_u = time_u;
	break;
      }	else if (encumbent > lgrg->wrong_clauses) {
	encumbent = lgrg->wrong_clauses;	
	enci = iter;
#if defined DEBUG2||DEBUG
	fprintf(stderr, "enc: %d at %d iter\n", encumbent,iter);
#endif
      }

      if (!udx) {
#ifdef DEBUG_DIFF
	printf ("---local min---\n");
#endif
	if (lgrg->enable_wrong_cl_list)
	  udu = update_u2(lgrg, MULTI*INC_U);
	else if (lgrg->diff_flag)
	  udu = update_u_diff(lgrg, MULTI*INC_U);
	else
	  udu = update_u(lgrg, MULTI*INC_U);  
      }
      else if  ((iter+1)%UPDATE_U_INT == 0 ) {
	if (lgrg->enable_wrong_cl_list)
	  udu = update_u2(lgrg, INC_U);
	else if (lgrg->diff_flag)
	  udu = update_u_diff(lgrg, INC_U);
	else
	  udu = update_u(lgrg, INC_U);
      }

      if (!lgrg->diff_flag  && iter >= lgrg->vars/3) {
	lgrg->diff_flag = 1;
	init_diff_struct(lgrg);
      }

#ifdef DEBUG
      if (iter%100 == 100) {
        times(&t_buffer);
        fprintf(stdout, "# iter %d, user time %.1f sec, iter/sec %.1f\n",
                iter+1,
                ((float) t_buffer.tms_utime - old_t_buffer.tms_utime)/60,
                (iter+1)*60*1.0/( t_buffer.tms_utime - old_t_buffer.tms_utime));
      }
#endif
    }
        
    avevalue += lgrg->wrong_clauses;

    if (iter == iterations) 
      fprintf(stdout, "Maximum of %d iterations reached with best %d unsatisfied clause(s) at %d iter.\n", iter, encumbent,enci);
    
    if (encumbent < best_obj)
      best_obj = encumbent;

    fprintf(stdout, "Final primary variable values (the first 30 variables): \n");
    for (i=0; i<lgrg->vars && i < 30; i++) {
      fprintf(stdout, "%d ", lgrg->prim_vars[i].value);
    }

    fprintf(stdout, "\nFinal Lagrange-multiplier values (the first 30 values): \n");
    for (i=0; i<lgrg->rows && i < 30; i++) 
      fprintf(stdout, "%d ", lgrg->dual_vars[i]);
    fprintf(stdout, "\n");

    
    set_init_points (lgrg, INIT_U);
    eval_h_value(lgrg);
    lgrg->cur_lgrg = eval_lagrangian(lgrg);
    lgrg->wrong_clauses = eval_wrong_clauses(lgrg);

    times(&t_buffer);
    fprintf(stdout, "In this round, user time %.3f sec, system time %.3f sec\n",
	    ((float) t_buffer.tms_utime - old_t_buffer.tms_utime)/60,
	    ((float) t_buffer.tms_stime - old_t_buffer.tms_stime)/60);
    
    /* fprintf(stdout, "%d Lagrangian evaluations.\n", NEVALS);*/

    old_t_buffer.tms_utime = t_buffer.tms_utime;
    old_t_buffer.tms_stime = t_buffer.tms_stime;
  }

  times(&t_buffer);
  if (count > 0) {
    fprintf(stdout, "\nBest solution in %d/%d successful rounds: %d unsatisfied clause(s). %d flips/sec\n", count, NROUND, best_obj, avefps/count);
    fprintf(stdout, "Iterations/successful round: ave %.1f, min %d, max %d\n", ((float) aveiter)/count, miniter, maxiter);
    fprintf(stdout, "Time/successful round (sec): ave %.3f (user), %.3f (system); min %.3f, max %.3f\n",
	  ((float) suc_time_u)/60/count,((float) suc_time_s)/60/count,((float) min_time_u)/60,((float) max_time_u)/60);
  } else {
    fprintf(stdout, "\nBest solution in %d/%d successful rounds: %d unsatisfied clause(s)\n", count, NROUND, best_obj);
  }

  fprintf(stdout, "Ave time/round (w/ initialization): user %.3f sec, system %.3f sec\n",
	  ((float) t_buffer.tms_utime)/60/NROUND,((float) t_buffer.tms_stime)/60/NROUND);
}

/* read the problem into the lgrg structure */
lgrgrec *lgrg_init (FILE *fp, short verbose)
{
  short Debug;
  lgrgrec *lgrg;
  int **tmp_aux;
  int i,j,k, count, max_len;
  int c, var;
  char kind[100];
  int *temp_in, NIN=2000000, index;
  int base, inc, temp;

  Debug = verbose;

  /* create lgrg structure */
  lgrg = (lgrgrec *) malloc (sizeof(lgrgrec));

  while (getchar()!='p'){
    while ((c = getchar()) != '\n'){
      if (c == EOF) { printf("Bad input\n"); exit(-1); }
    }
  }

  scanf("%s", kind);
  if (strcmp(kind, "cnf")) {
    fprintf(stderr, "Bad input\n"); 
    exit(-1);
  }

  lgrg->vars = getint();
  lgrg->rows = getint();

  MAX_IMPR_LIST = (int) lgrg->vars/1.5;
  /* AUX_COLS = (30*lgrg->rows/lgrg->vars >100) ? 30*lgrg->rows/lgrg->vars:300; */
    AUX_COLS = 1000;

  lgrg->aux_cols = AUX_COLS;

  temp_in = (int *) malloc (NIN* sizeof(int));
  /* scan once to get 'nlit' for memory allocation */
  index = 0;
  for (count = i=0; i<lgrg->rows; i++) {
      while (temp_in[index++] = getint()) {
        count++;
      }
      if (index > NIN-100) {
	printf("Error: Length of temporary storage of %d is not large enough to
  hold the problem\n", NIN);
	free(temp_in);
	exit(-1);
      }
  }
  /* fprintf(stderr,"nlits %d\n", count);
  fprintf(stderr,"nclause %d\t total len %d\n", lgrg->rows, index); */
  lgrg->nlits = count;

  lgrg->body = (int *) malloc ((lgrg->rows+lgrg->nlits+1)*sizeof(int));
  lgrg->row_base = (int *) malloc ((lgrg->rows)*sizeof(int));

  /* scan the 2nd times to read data */
  base = 1;
  index = 0;
  for (count=i=0; i<lgrg->rows; i++) {
    lgrg->row_base[i] = base;
    while (temp=temp_in[index++]) {
      lgrg->body[base+count+1]= temp;
      count++;
    }
    lgrg->body[base]= count;
    base += count+1;
    count = 0;
  }
  free(temp_in);

#ifdef DEBUG
  printf("The data set\n");
  for (count=i=0; i<lgrg->rows; i++) {
    base = lgrg->row_base[i];
    temp = lgrg->body[base];
    for (j = 0; j<temp; j++)
      printf("%d  ", lgrg->body[base+j+1]);
    printf("\n");
  }
#endif

  tmp_aux = (int **) malloc ((lgrg->vars+1)*sizeof(int *));
  for (i=0; i<=lgrg->vars; i++) {
    tmp_aux[i] = (int *) malloc ((lgrg->aux_cols)*sizeof(int));
    tmp_aux[i][0] = 1;
  }

  for (i=0; i<lgrg->rows; i++) {
    base = lgrg->row_base[i];
    temp = base+1+lgrg->body[base];
    for (j=base+1; j<temp; j++) {
	/* fill in tmp_aux */
	var = abs(lgrg->body[j]);
	if (tmp_aux[var][0] == lgrg->aux_cols-1) {
	  fprintf(stdout, "width %d of auxilary matrix is not large enough for var %d.\n",lgrg->aux_cols, var);
	  exit (-1);
	} else {
	  tmp_aux[var][tmp_aux[var][0]] = i;
	  tmp_aux[var][0]++;
	}
      }
  }

  for (i=1, max_len=0; i<=lgrg->vars; i++) {
    if (tmp_aux[i][0] > max_len) 
      max_len = tmp_aux[i][0];
  }
  /* fprintf(stderr,"Max_col_len_of_aux: %d\n", max_len); */


  lgrg->aux = (int *) malloc ((lgrg->vars+lgrg->nlits+1)*sizeof(int));
  lgrg->aux_base = (int *) malloc ((lgrg->vars+1)*sizeof(int));
  base = 1;
  index = 0;
  for (i=0; i<=lgrg->vars; i++) {
    lgrg->aux_base[i] = base;
    lgrg->aux[base] = tmp_aux[i][0] - 1;
    for (j=1; j < tmp_aux[i][0]; j++) 
	lgrg->aux[base+j] = tmp_aux[i][j];
    base += tmp_aux[i][0];
  }
 
  for (i=0; i<=lgrg->vars; i++)
    free(tmp_aux[i]);
  free(tmp_aux);


  lgrg->diff_flag = 0;
  lgrg->prim_vars = (var_str *) malloc ((lgrg->vars)*sizeof(var_str));
  lgrg->impr_list = (int *) malloc (MAX_IMPR_LIST * sizeof(int));
  
  lgrg->dual_vars = (REAL *) malloc ((lgrg->rows)*sizeof(REAL));
  lgrg->h = (char *) malloc ((lgrg->rows)*sizeof(char));

  lgrg->infinite = 0x7fffffff;
  lgrg->enable_wrong_cl_list = 0;

  set_init_points (lgrg, INIT_U);

  eval_h_value(lgrg);
  lgrg->cur_lgrg = eval_lagrangian(lgrg);
  lgrg->wrong_clauses = eval_wrong_clauses(lgrg);

  lgrg->descent_f = 0; /* hillclimbing */
  lgrg->descent_f = DESC_F; /* steepest hillclimbing */
  lgrg->descent_o = DESC_O;

  if (Debug) 
    print_lgrg(lgrg, stdout);

  return lgrg;
}


/* print lgrg information for debug */
void print_lgrg(lgrgrec *lgrg, FILE *output) 
{
  int i,j;
  int base, temp;

  fprintf(output, "nvars %d, nclauses %d\n", lgrg->vars, lgrg->rows);
  for (j=0; j<lgrg->rows; j++) {
    temp = lgrg->body[base];
    for (i=0; i< temp; i++) 
      fprintf(output, "%d ", lgrg->body[base+i+1]);
    fprintf(output,"\n");
    base += temp+1;
  }
  
  print_initial(lgrg, output);
}

void set_init_points (lgrgrec *lgrg, int ini_u)
{
  int i;
  extern double drand48();

  if (NROUND ==1) { /* always starts from 0 */
    for (i=0; i<lgrg->vars; i++)
      lgrg->prim_vars[i].value = 0;
  } else {
    for (i=0; i<lgrg->vars; i++)
      lgrg->prim_vars[i].value = rint(drand48());
  }

  for (i=0; i<lgrg->rows; i++)
    lgrg->dual_vars[i] = ini_u;
}
  

/* print inital values */
void print_initial (lgrgrec *lgrg, FILE *output)
{
  int i;
  fprintf(output, "\nInitial values --- Primary vars: ");
  for (i=0; i<lgrg->vars; i++)
    fprintf(output, " %d", lgrg->prim_vars[i].value);
  fprintf(output,"\n");
}


/* given primary variable values x,
   return function value h.

   h(x):  h(x) =  0       if satisfied
               =  1       if not
*/
void eval_h_value(lgrgrec* lgrg)
{
  int i,j, tmp;
  int base, temp;

  for (j=0; j< lgrg->rows; j++) {
    base = lgrg->row_base[j];
    temp = base+1+lgrg->body[base];
    lgrg->h[j] = 1;
    for (i=base+1; i<temp; i++) {
      tmp = lgrg->body[i];
      if ( tmp > 0) {
	if (lgrg->prim_vars[tmp-1].value) {
	  lgrg->h[j] = 0;
	  break;
	}
      } else if (tmp < 0) {
	if (!lgrg->prim_vars[-tmp-1].value) {
	  lgrg->h[j] = 0;
	  break;
	}
      } else
	break;
    }
  }
}

 
/* given primary variable values x and dual variables u,
   return lagrangian function value l.

   l (x, u) = u h(x)
*/
REAL eval_lagrangian(lgrgrec* lgrg)
{
  REAL l;
  int i,j;

  NEVALS ++;

  l = 0;

  for (i=0; i<lgrg->rows; i++) 
    l += lgrg->dual_vars[i] * lgrg->h[i]; 
    
  return l;
}

/* the number of wrong clauses based on current assignment */
REAL eval_wrong_clauses(lgrgrec* lgrg)
{
  REAL l;
  int i,j;

  l = 0;
  for (i=0; i<lgrg->rows; i++) 
    if (lgrg->h[i])
      l ++; 
  return l;
}

/* given primary variable values x and dual variables u,
   update x along the gradient direction.

*/
int update_x(lgrgrec* lgrg)
{
  int i,j;
  int update;

  if (lgrg->descent_o != 1) {
    fprintf(stderr, "Descent order of %d has not been implemented.\n",
	    lgrg->descent_o);
    exit(-1);
  }

  if (lgrg->descent_f == 0)  {      /* stochastic hill-climbing */
    update = update_x_sto(lgrg);
  }
  else
    update = update_x_det(lgrg);

  return update;
}


/* update x in steep-descent manner */
int update_x_det(lgrgrec* lgrg)
{
  int i,j, k;
  REAL orig_l, current_l, best_l, best_v, save_value;
  int best_i, update;
  short *mark_l, count_mark, rec_mark[3], flag;

  mark_l = (short *) malloc (lgrg->columns * sizeof(short));
  for (i=0; i<lgrg->columns; i++) mark_l[i]=0;

  /* this is only for order of 1 */
  update = pure01_1bit_best(lgrg);

  return update;
}

/* update x in stochastic descent manner */
int update_x_sto(lgrgrec* lgrg)
{
  int i,j, update;
  
  if (lgrg->enable_wrong_cl_list || lgrg->wrong_clauses <= 0) {
/*    update = pure01_1bit_best(lgrg); */
    if (!lgrg->enable_wrong_cl_list) {
      lgrg->enable_wrong_cl_list = 1;
      get_wrong_cl_list (lgrg);
    }
    update = pure01_1bit_hillclimb2(lgrg);
  }
  else if (lgrg->diff_flag) {
    if (!diff_wc_flag && lgrg->wrong_clauses < /*100 */lgrg->rows/200) {
      diff_wc_flag = 1;
      get_wrong_cl_list (lgrg);
    }      
    update = pure01_1bit_hillclimb_diff(lgrg);
  }
  else
    update = pure01_1bit_hillclimb(lgrg);

  return update;
}

/* update x in stochastic descent manner */
int update_x_sto2(lgrgrec* lgrg)
{
  int i,j, update;
  int *rand_list;

  rand_list = (int *) malloc (lgrg->vars * sizeof(int));
  rand_perm(lgrg->vars, rand_list);

  /* this is only for order of 1 */
  update = pure01_1bit_hillclimb_sto(lgrg, rand_list);
  free(rand_list);

  return update;
}


/* given primary variable values x and dual variables u,
   update u along the h direction.

*/
int update_u(lgrgrec* lgrg, int inc)
{
  int i;

  if (UPBOUND_U) {
    for (i=0; i<lgrg->rows; i++) 
      if (lgrg->h[i] && lgrg->dual_vars[i]<UPBOUND_U) {
	lgrg->dual_vars[i] ++; 
      }
  } else
    for (i=0; i<lgrg->rows; i++) 
/*      if (lgrg->h[i]) {
	lgrg->dual_vars[i] ++; 
	lgrg->cur_lgrg ++;
      }
*/
      if (lgrg->h[i]) {
	lgrg->dual_vars[i] += inc; 
	lgrg->cur_lgrg += inc;
      }
  return 1;
}

/* take advantage of the wrong_cl_list */
int update_u2(lgrgrec* lgrg, int inc)
{
  int i, k, key;

  for (k=0; k<lgrg->wrong_clauses; k++) {
    key = lgrg->wrong_cl_list[k]; /* the clause */
    lgrg->dual_vars[key] += inc; 
    lgrg->cur_lgrg += inc;
  }
  return 1;
}
    

int update_u_diff(lgrgrec* lgrg, int inc)
{
  int i, j, k;
  int tmp, ntmp;
  int base, temp;

  if (diff_wc_flag) {
   while (lgrg->n_impr_list == 0) 
    for (k=0; k<lgrg->wrong_clauses; k++) {
      i = lgrg->wrong_cl_list[k]; /* the clause */
      base = lgrg->row_base[i];
      ntmp = base+1+lgrg->body[base];
      for (j=base+1; j<ntmp; j++) {
	tmp = lgrg->body[j];
	tmp = tmp>0? tmp-1: -tmp-1;
	  
	if (lgrg->prim_vars[tmp].diff <=0 && inc+lgrg->prim_vars[tmp].diff >0)
	  insert_impr_list(lgrg, tmp);
	lgrg->prim_vars[tmp].diff += inc;  /* impr */
      }
      
      lgrg->dual_vars[i] += inc; 
      lgrg->cur_lgrg += inc;
    }
    return 1;
  }

    
  while (lgrg->n_impr_list == 0) {
    for (i=0; i<lgrg->rows; i++) 
      if (lgrg->h[i]) {
	base = lgrg->row_base[i];
	ntmp = base+1+lgrg->body[base];
	for (j=base+1; j<ntmp; j++) {
	  tmp = lgrg->body[j];
	  tmp = tmp>0? tmp-1: -tmp-1;
	  
	  if (lgrg->prim_vars[tmp].diff <=0 && inc+lgrg->prim_vars[tmp].diff >0)
	    insert_impr_list(lgrg, tmp);
	  lgrg->prim_vars[tmp].diff += inc;  /* impr */
	}

	lgrg->dual_vars[i] += inc; 
	lgrg->cur_lgrg += inc;
      }
  }

  return 1;
}

/* generate a random permutation of numbers
   0, 1, ..., n-1
*/

void rand_perm(int n, int *list)
{
  int i,j,a;
  extern double drand48();

  for (i=0; i<n; i++) list[i] = -1;

  for (i=0; i<n; i++) {
    a = floor(drand48()*n);
    if (list[a] < 0)
      list[a] = i;
    else {
      for (j = a+1;1; j++) {
	if (j >= n) j = 0;
	if (list[j] < 0) {
	  list[j] = i;
	  break;
	}
	if (j==a) {
	  fprintf(stderr, "infinite loop!\n");
	  exit(-1);
	}
      }
    }
  }

/* for debug 
  printf("Random list: ");
  for (i=0; i<n; i++) 
    printf(" %d ", list[i]);
  printf("\n");
*/
}

/* for pure 0-1 IP problems, find the best in 1-bit neighbourhood.
   input: x, u
	   return update;
	   
*/
int pure01_1bit_best(lgrgrec* lgrg)
{
  int i, j, k;
  REAL old_l, new_l, best_l;
  int start;
  int update, best_i, clause, temp, old_wrong;
  int flag;
  int count = 0;
  int base, temp2;

  update = 0; 
  old_l = best_l = lgrg->cur_lgrg;

  for (i=0; i<lgrg->vars; i++) {
    flag = 0;
    base = lgrg->aux_base[i+1];
    temp = base+1+lgrg->aux[base];
    for (j=base+1; j<temp; j++) {   
      clause = lgrg->aux[j];
      if (lgrg->h[clause]) {
	flag = 1;
	break;
      }
    }
    if (!flag)
      continue;
  
    new_l = old_l;
    /* flip one bit of x */
    for (j=base+1; j<temp; j++) {   
      clause = lgrg->aux[j];
      new_l += (eval_clause_1bit(lgrg, clause, i+1) - lgrg->h[clause]) *
	lgrg->dual_vars[clause];
    }

    if (new_l < best_l) {
      best_l = new_l;
      best_i = i;
    }
  }

  if (best_l == old_l)
    return 0;
  else {
    lgrg->cur_lgrg = new_l;
    old_wrong = lgrg->wrong_clauses;
    base = lgrg->aux_base[best_i+1];
    temp2= base+1+lgrg->aux[base];
    for (j=base+1; j<temp; j++) { 
      clause = lgrg->aux[j]; 
      temp = eval_clause_1bit(lgrg, clause, best_i+1);
      lgrg->wrong_clauses += temp - lgrg->h[clause];

      lgrg->h[clause] = temp;
    }
    lgrg->prim_vars[best_i].value = 1-lgrg->prim_vars[best_i].value;
  }
} 

int pure01_1bit_hillclimb_diff(lgrgrec* lgrg)
{
  int i, j, var, index;
  extern double drand48();

  if (lgrg->n_impr_list == 0)
    return 0;

#ifdef DEBUG_DIFF
  printf("before flip: ");
  print_impr_list(lgrg);
#endif

/*
  index = random_0_to(lgrg->n_impr_list);
  var = lgrg->impr_list[index];
*/
  var = lgrg->impr_list[0];

#ifdef DEBUG_DIFF
 printf("deb: index %d, var %d, n_impr_list %d\n", index, var, lgrg->n_impr_list); 
#endif

  flip_update_diff(lgrg, var);  

#ifdef DEBUG_DIFF
  printf("\nAfter flip %d: ", var);
  print_impr_list(lgrg);
#endif

  if (lgrg->n_impr_list == 0)
    return 0;
}

/* for SAT problems, find the first improvement in 1-bit neighbourhood.
   input: x, u, h (in lgrg), cur_l
   output: new x, new h
	   return update;
*/
int pure01_1bit_hillclimb(lgrgrec* lgrg)
{
  int i, j, k;
  REAL old_l, new_l;
  int start;
  int update, best_i, clause, temp, old_wrong;
  int flag, side_f, side_i, side_l;
  int count = 0;
  float dir;
  extern double drand48();
  int base, temp2;

  side_f= 1;
  update = 0; 
  old_l = lgrg->cur_lgrg;

#ifdef SAVE_P
  i = start = SAVE_POINT;
  for (k=0; k<lgrg->vars; k++) {
    i = (i+1)%lgrg->vars;
#else
  dir = drand48();
  i = start = rint(drand48() * lgrg->vars);
  for (k=0; k<lgrg->vars; k++) {
    if (dir >0.5)
      i = (i+lgrg->vars-1)%lgrg->vars;
    else
      i = (i+1)%lgrg->vars;
#endif
  
    base = lgrg->aux_base[i+1];
    temp2 = base+1+lgrg->aux[base];
  
    if (WRONG_VAR_MODE) {
      flag = 0;
      for (j=base+1; j<temp2; j++) {   
	clause = lgrg->aux[j];
	if (lgrg->h[clause]) {
	  flag = 1;
	  break;
	}
      }
      if (!flag)
	continue;
    }

    new_l = old_l;
    /* flip one bit of x */
    for (j=base+1; j<temp2; j++) {   
      clause = lgrg->aux[j];
      if (lgrg->h[clause])
	new_l += -lgrg->dual_vars[clause];
      else if (eval_clause_1bit(lgrg, clause, i+1))
	new_l += lgrg->dual_vars[clause];
    }
    
    if (new_l < old_l) {
      update = 1;
      lgrg->cur_lgrg = new_l;
      old_wrong = lgrg->wrong_clauses;
      for (j=base+1; j<temp2; j++) { 
	clause = lgrg->aux[j]; 
	if (lgrg->h[clause]) {
	  lgrg->wrong_clauses--;
	  lgrg->h[clause] = 0;
	} else if (eval_clause_1bit(lgrg, clause, i+1)) {
	  lgrg->wrong_clauses ++;
	  lgrg->h[clause] = 1;
	}
      }
      lgrg->prim_vars[i].value = 1-lgrg->prim_vars[i].value;

#ifdef SAVE_P
      SAVE_POINT = i+1;
#endif

#ifdef DEBUG
      printf("  after %d flips, ", k+1);
      printf("fliping var %d from %d gives lgrg %d and wrong bit %d impr\n", i+1, 1-lgrg->prim_vars[i].value, old_l-new_l, old_wrong - lgrg->wrong_clauses );
#endif     
      break;    
    }
#ifdef SIDEWALK
    else if (side_f && new_l == old_l) {
      side_f = 0;
      side_i = i;
    }
#endif

  }

#ifdef SIDEWALK
  if (!update && !side_f) {
/*    update = 1; */
    old_wrong = lgrg->wrong_clauses;
    base = lgrg->aux_base[side_i+1];
    temp2 = base+1+lgrg->aux[base];
    for (j=base+1; j<temp2; j++) { 
      clause = lgrg->aux[j]; 
      temp = eval_clause_1bit(lgrg, clause, side_i+1);
      lgrg->wrong_clauses += temp - lgrg->h[clause];
      
      lgrg->h[clause] = temp;
    }
    lgrg->prim_vars[side_i].value = 1-lgrg->prim_vars[side_i].value;
#ifdef SAVE_P
      SAVE_POINT = side_i+1;
#endif

  }
#endif    
    
  return update;
}

/* for SAT problems, find the best improvement for the first wrong clauses
   in 1-bit neighbourhood.
	   return update;
*/
int pure01_1bit_hillclimb2(lgrgrec* lgrg)
{
  int i, j, k, index;
  int update, best_i, clause, temp, old_wrong;
  int impr, maximpr, maximprx, x, key;
  extern double drand48();
  int base, ntmp, base_aux, ntmp_aux;

  update = 0; 
  maximpr = 10000000;
/*
#ifdef SAVE_P
  index = SAVE_POINT;
#else
  index = rint(drand48() * lgrg->wrong_clauses);
#endif
*/

  index = rint(drand48() * lgrg->wrong_clauses);
  for (k=0; k<lgrg->wrong_clauses; k++) {
    index = (index+1)%lgrg->wrong_clauses;
    key = lgrg->wrong_cl_list[index];     /* the clause */
    base = lgrg->row_base[key];
    ntmp = base+1+lgrg->body[base];
    for (i=base+1; i<ntmp; i++) {
      x = abs(lgrg->body[i]);
      impr = 0;
      if(!x)	break;
      /* flip one bit of x */
      base_aux = lgrg->aux_base[x];
      ntmp_aux = base_aux + 1 + lgrg->aux[base_aux];
      for (j=base_aux+1; j<ntmp_aux; j++) {   
	clause = lgrg->aux[j];
	if (lgrg->h[clause])
	  impr += -lgrg->dual_vars[clause];
	else if (eval_clause_1bit(lgrg, clause, x))
	  impr += lgrg->dual_vars[clause];
      }
    
      if (impr < 0) 
	break;

      if (impr < maximpr) {
	maximpr = impr;
	maximprx = x;
      }
    }

    if (impr < 0) 
      break;
  }

  if (impr < 0) {
    lgrg->cur_lgrg += impr;
    old_wrong = lgrg->wrong_clauses;
    base_aux = lgrg->aux_base[x];
    ntmp_aux = base_aux + 1 + lgrg->aux[base_aux];
    for (j=base_aux+1; j<ntmp_aux; j++) {
      clause = lgrg->aux[j];
      if (lgrg->h[clause]) {
	remove_wrong_cl_list(lgrg, clause);
	lgrg->wrong_clauses --;
	lgrg->h[clause] = 0;
      } else if (eval_clause_1bit(lgrg, clause, x)) {
	  lgrg->wrong_clauses ++;
	  lgrg->h[clause] = 1;
	  insert_wrong_cl_list(lgrg, clause);
	}
    }
    lgrg->prim_vars[x-1].value = 1-lgrg->prim_vars[x-1].value;

/*
#ifdef SAVE_P
      SAVE_POINT = key;
#endif
*/
    return 1;
  } 

  if (UPDATE_U_MODE ==1) {
    update_u2(lgrg, maximpr+2);
#ifdef DEBUG
    printf("local min, increase u by %d\n", maximpr+2);
#endif
    return 1;
  }


#ifdef SIDEWALK
  if (maximpr > 0)
    return 0;

  lgrg->cur_lgrg += maximpr;
    old_wrong = lgrg->wrong_clauses;
    base_aux = lgrg->aux_base[maximprx];
    ntmp_aux = base_aux+1+lgrg->aux[base_aux];
    for (j=base+1; j<ntmp_aux; j++) { 
      clause = lgrg->aux[j]; 
      if (lgrg->h[clause]) {
	remove_wrong_cl_list(lgrg, clause);
	lgrg->wrong_clauses --;
	lgrg->h[clause] = 0;
      } else if (eval_clause_1bit(lgrg, clause, maximprx)) {
	  lgrg->wrong_clauses ++;
	  lgrg->h[clause] = 1;
	  insert_wrong_cl_list(lgrg, clause);
	}
    }
    lgrg->prim_vars[maximprx-1].value = 1-lgrg->prim_vars[maximprx-1].value;
#endif

  return 0;
}

    


/* for SAT problems, find the first improvement in 1-bit neighbourhood.
   input: x, u, h (in lgrg), cur_l
   output: new x, new h
	   return update;
*/
int pure01_1bit_hillclimb_sto(lgrgrec* lgrg, int *rand_list)
{
  int i, j, k;
  REAL old_l, new_l;
  int update, best_i, clause;
  int count = 0;
  int base, ntmp;

  update = 0; 
  old_l = lgrg->cur_lgrg;

  for (k=0; k<lgrg->vars; k++) {
    i = rand_list[k];
    new_l = old_l;
    /* flip one bit of x */
    base = lgrg->aux_base[i+1];
    ntmp = base+1+lgrg->aux[base];
    for (j=base+1; j<ntmp; j++) {   
      clause = lgrg->aux[j];
      new_l += (eval_clause_1bit(lgrg, clause, i+1) - lgrg->h[clause]) *
	lgrg->dual_vars[clause];
    }
    
    if (new_l < old_l) {
      update = 1;
      lgrg->cur_lgrg = new_l;
      for (j=base+1; j<ntmp; j++) {   
	clause = lgrg->aux[j];
	lgrg->h[clause] = eval_clause_1bit(lgrg, clause, i+1);
      }
      lgrg->prim_vars[i].value = 1-lgrg->prim_vars[i].value;
      
      break;
    }
  }
  return update;
}


int getint()
{
    int x;
    if (scanf("%d", &x)!=1){
	fprintf(stderr, "Bad input\n"); 
	exit(-1);
    }
    return x;
}

/* evaluate the n'th clause with the m'th variable fliped
   return whether the clause is satisfied (0) or not (1).
*/
int eval_clause_1bit(lgrgrec *lgrg, int n, int m)
{
  int i;
  int base, ntmp;

  base = lgrg->row_base[n];
  ntmp = base+1+lgrg->body[base];
  for (i=base+1; i<ntmp; i++) {
    if (lgrg->body[i]> 0) {
      if (lgrg->body[i]== m) {
	if (!lgrg->prim_vars[m-1].value) 
	  return 0;
      } else if (lgrg->prim_vars[lgrg->body[i]-1].value)
	return 0;
    } else if (lgrg->body[i] < 0) {
      if (-lgrg->body[i] == m) {
	if (lgrg->prim_vars[m-1].value) 
	  return 0;
      } else if (!lgrg->prim_vars[-lgrg->body[i]-1].value)
	return 0;
    } else
      return 1;
  }
  return 1;
}
  
 
void check_lgrg(lgrgrec* lgrg)
{
  int i, j, tmp, value;
  int base, ntmp;
  
  for (j=0; j< lgrg->rows; j++) {
    value = 1;
    base = lgrg->row_base[j];
    ntmp = base+1+lgrg->body[base];
    for (i=base+1; i<ntmp; i++) {
      tmp = lgrg->body[i];
      if ( tmp > 0) {
	if (lgrg->prim_vars[tmp-1].value) {
	  value = 0;
	  break;
	}
      } else if (tmp < 0) {
	if (!lgrg->prim_vars[-tmp-1].value) {
	  value = 0;
	  break;
	}
      } else
	break;
    } 
    if (value != lgrg->h[j]) {
      printf("Something wrong in evaluate h\n");
      exit (-1);
    }
  }
}

void get_wrong_cl_list (lgrgrec *lgrg)
{
  int i,j;

  for (i=j=0; j< lgrg->rows; j++) {
    if (lgrg->h[j]) {
      lgrg->wrong_cl_list[i] = j;
      i++;
      if (i>wrong_cl_list_l) {
	printf("Error: the wrong clause list,  wrong_cl_list_l =  %d, defined
  in lgrg_v2.h is not long enough.\n", wrong_cl_list_l);
	exit (-1);
      }
    }
  }
}

void remove_wrong_cl_list(lgrgrec* lgrg, int clause)
{
  int i, j;

#ifdef DEBUG
  printf("remove %d from w.c.l (wrong_clauses %d)\n", clause, lgrg->wrong_clauses);
#endif

  for (i=0; i<lgrg->wrong_clauses; i++)
    if (lgrg->wrong_cl_list[i] == clause) {
      for (j=i; j<lgrg->wrong_clauses-1; j++) 
	lgrg->wrong_cl_list[j] = lgrg->wrong_cl_list[j+1];
      break;
    }

  if (i >= lgrg->wrong_clauses) {
    printf("clause %d is not in wrong_cl_list!\n", clause);
    exit(-1);
  }
}
    
void insert_wrong_cl_list(lgrgrec* lgrg, int clause)
{
  int i, j;

#ifdef DEBUG
  printf("insert %d in w.c.l (wrong_clauses %d)\n", clause, lgrg->wrong_clauses);
#endif
  if (lgrg->wrong_clauses >= wrong_cl_list_l) {
    printf("Wrong: wrong_cl_list is not long enough\n");
    exit(-1);
  }
  lgrg->wrong_cl_list[lgrg->wrong_clauses-1] = clause;
}
    


/* the following procedures are for the "diff" structure implementation */
void init_diff_struct (lgrgrec *lgrg)
{
  int i, j, tmp, var;
  int nops;
  int save_v, ntmp;
  int base;

  for (i=0; i<lgrg->vars; i++) 
    lgrg->prim_vars[i].diff = 0;

  /* fill in prim_vars[i].diff */
  for (i=0; i<lgrg->rows; i++) {
    base = lgrg->row_base[i];
    ntmp = base+1+lgrg->body[base];
    if (lgrg->h[i]) {
      for (j=base+1; j<ntmp; j++) {
	tmp = lgrg->body[j];
	if (tmp < 0)
	  tmp = -tmp;
	
	lgrg->prim_vars[tmp-1].diff += lgrg->dual_vars[i];  /* impr */
      }
    } else {
      /* if nops > 1:  do nothing;
	 if nops = 1:  the var is critical, flipping it reduce its diff;
	               others do not change.
	 if nops = 0 (h=0):  all vars increase diff by flipping.
      */

      nops = 0;
      for (j=base+1; j<ntmp; j++) {
	tmp = lgrg->body[j];

	if (make(tmp)) {
	  save_v = (tmp>0)? tmp:-tmp;
	  nops++;
	}
      }
      if (nops == 1)
	lgrg->prim_vars[save_v-1].diff -= lgrg->dual_vars[i];  /* degr */
    }
  }

  /* fill in impr_list */
  for (i=j=0; i<lgrg->vars; i++) 
    if (lgrg->prim_vars[i].diff > 0) {
      lgrg->impr_list[j] = i;
      j++;
      if (j > MAX_IMPR_LIST) {
	printf("The impr_list is not long enough!");
	exit(-1);
      }
    }
  lgrg->n_impr_list = j;
}
	
/* flipping x and 
   update 'diff' values for all vars affected by flipping var x */
void flip_update_diff(lgrgrec *lgrg, int var)
{
  int i, j, x, clause, k;
  int nops, var_make, save_v;
  int get, loss;
  int ntmp, difftmp;
  int base, base_aux, ntmp_aux;

  get = loss = 0;

  base_aux = lgrg->aux_base[var+1];
  ntmp_aux = base_aux+1+lgrg->aux[base_aux];
  for (i=base_aux+1; i<ntmp_aux; i++) {
    clause = lgrg->aux[i];
    base = lgrg->row_base[clause];
    ntmp = base+1+lgrg->body[base];

    if (lgrg->h[clause]) {    /* unsat clause */
      lgrg->h[clause] = 0;    /* sat now */
      if (diff_wc_flag) {	
	remove_wrong_cl_list(lgrg, clause);
	lgrg->wrong_clauses --;
      }
      get ++;  
#ifdef DEBUG_DIFF
printf("get c%d  ",clause);
#endif
      for (j=base+1; j<ntmp; j++) {
	x = lgrg->body[j];
	if (x<0) x = -x;

	  
	if (x-1 != var) {
	  if (lgrg->prim_vars[x-1].diff >0 && 
	      lgrg->prim_vars[x-1].diff <= lgrg->dual_vars[clause]) {
	    remove_impr_list(lgrg, x-1);
	  }
	  lgrg->prim_vars[x-1].diff -= lgrg->dual_vars[clause];  
	  /* all other vars' "diff" decrease because this clause becomes sat */
	}
      }
    } else {     /* sat before */
      nops = var_make = 0;
      for (j=base+1; j<ntmp; j++) {
	x = lgrg->body[j];
	if (abs(x) == (var+1)) {
	  if (make(x)) {
	    var_make = 1;
	    nops ++;
	  }
	} else if (make(x))  {      /* other vars */
	  save_v = abs(x);
	  nops ++;
	}
      }

      if (nops == 2 && var_make) { /* save_v becomes critical */
	if (lgrg->prim_vars[save_v-1].diff >0 && 
	    lgrg->prim_vars[save_v-1].diff <= lgrg->dual_vars[clause]) {
	  remove_impr_list(lgrg, save_v-1);
  	}
	lgrg->prim_vars[save_v-1].diff -= lgrg->dual_vars[clause];
      }
      else if (nops == 1 && !var_make) { /* save_v becomes non-critical */
	if (lgrg->prim_vars[save_v-1].diff <=0 && 
	    -lgrg->prim_vars[save_v-1].diff < lgrg->dual_vars[clause])
	  insert_impr_list(lgrg, save_v-1);
	lgrg->prim_vars[save_v-1].diff += lgrg->dual_vars[clause];
      }
      else if (nops ==1 && var_make)  { /* the clause is unsat now, 
					 all vars increase diff */
	lgrg->h[clause] = 1;
	if (diff_wc_flag) {
	  lgrg->wrong_clauses ++;
	  insert_wrong_cl_list(lgrg, clause);
	}
	loss++;
#ifdef DEBUG_DIFF
printf("loss c%d  ",clause);
#endif
	for (j=base+1; j<ntmp; j++) {
	  x = lgrg->body[j];
	  x = (x<0)? -x-1: x-1;
 
	  if (x != var) {
	    if (lgrg->prim_vars[x].diff <=0 && 
		-lgrg->prim_vars[x].diff < lgrg->dual_vars[clause])
	      insert_impr_list(lgrg, x);
	    lgrg->prim_vars[x].diff += lgrg->dual_vars[clause];
	  }
	}
      }
    }
  }

  /* flip var */
  if (!diff_wc_flag) lgrg->wrong_clauses -= get-loss;
  lgrg->cur_lgrg -= lgrg->prim_vars[var].diff;
  lgrg->prim_vars[var].value = (lgrg->prim_vars[var].value)? 0:1;
  lgrg->prim_vars[var].diff *=-1;
  remove_impr_list(lgrg, var);
}
	 
	
/* insert var into impr_list */
void insert_impr_list(lgrgrec *lgrg, int var) {
  if (lgrg->n_impr_list == MAX_IMPR_LIST) {
    printf("When insert %d, the length of impr_lis exceeds maximum!", var);
    exit(-1);
  }

  lgrg->impr_list[lgrg->n_impr_list] = var;
  lgrg->n_impr_list++;
}


/* remove var from impr_list */
void remove_impr_list(lgrgrec *lgrg, int var) {
  int i;

  if (lgrg->n_impr_list == 0) {
    printf("When remove %d, the  impr_lis has nothing!", var);
    exit(-1);
  }

  for (i=0; i<lgrg->n_impr_list; i++) 
    if (lgrg->impr_list[i] == var)
      break;

  if (i == lgrg->n_impr_list) {
    printf("%d is not in impr_list\n", var);
    print_impr_list(lgrg);
    exit(-1);
  }

  for (; i<lgrg->n_impr_list-1; i++)
    lgrg->impr_list[i] = lgrg->impr_list[i+1];

  lgrg->n_impr_list--;
}


/* for debug, print vars in impr_list and their diff */
void  print_impr_list(lgrgrec * lgrg) 
{
  int i, k, var;

  printf("# is %d,  ", lgrg->n_impr_list);
  for (i=0; i<lgrg->n_impr_list; i++) {
    var = lgrg->impr_list[i];
    printf("%d (%d,%d)  ", var, lgrg->prim_vars[var].value, 
	   lgrg->prim_vars[var].diff);
    if (i%10==9)
      printf("\n");
  }

  printf("\n");
  printf("  w.c.l %d, lgrg %d\n", lgrg->wrong_clauses, lgrg->cur_lgrg);

for ( k=0; k < lgrg->vars; k++) {
  printf("%d ", lgrg->prim_vars[k].diff);
  if (k%10 == 0) printf("\n");
}
printf("\n");
}

int random_0_to(int n)  /* reture random number 1 .. n */
{
    long r;

      r = lrand48();
    return ((r >> 9) % n);
}
   
